from .hl7_worklist_parser import Hl7WorklistParser
import typing


class Hl7WorklistParserAssistovet(Hl7WorklistParser):

    def __init__(self, specific_fields: dict = None, patient_name_components_count: int = 5):
        super(Hl7WorklistParserAssistovet, self).__init__(specific_fields, patient_name_components_count)

    def parse(self, hl7_message: str) -> typing.Dict:
        # let's parse the default fields
        values = Hl7WorklistParser.parse(self, hl7_message=hl7_message)

        # let's add the field specific to Assistovet
        if values.get('PatientSpeciesDescription') is not None and values.get('PatientBreedDescription') is not None:
            # this should be an animal, so let's fill the responsible name
            last_name = self._get('PID.F5.R1.C1', default_value="")
            first_name = values.get('PatientMotherBirthName')
            if first_name is None: first_name = ""
            values['ResponsiblePerson'] = '^'.join([last_name, first_name])
        return values



from .hl7_worklist_parser import Hl7WorklistParser
import typing


class Hl7WorklistParserVetera(Hl7WorklistParser):

    def __init__(self, specific_fields: dict = None, patient_name_components_count: int = 5):

        vetera_dict = {
            'PatientID': 'PID.F2',
            'PatientSpeciesDescription': 'PID.F35',
            'PatientBreedDescription': 'PID.F36',
            'PatientSexNeutered': 'PID.F37',
            'BreedRegistrationNumber': 'PID.F38',
            '_scheduledProcedureStepStartDateTime': 'OBR.F6',
            'Modality': 'OBR.F21',
            'RequestingPhysician': 'OBR.F32',
            'RequestedProcedureDescription': 'OBR.F4'
        }
        if specific_fields is not None:
            vetera_dict.update(specific_fields)

        super(Hl7WorklistParserVetera, self).__init__(vetera_dict, patient_name_components_count)



import sys
import datetime
import paramiko
import time, os
import argparse
import logging
import schedule
import subprocess
logger = logging.getLogger(__name__)

class PostgresDumper:
    """
    Runs every day to create a gzip compressed dump of the postgres DB and write it to the destination (currently, only sftp)

    Warning:
    `postgresql-client` has to be installed before the execution of this script

    To restore:
    pg_restore --clean -U postgres -h 172.21.0.3 -p 5432 -d postgres Friday
    """
    def __init__(self, pg_host: str, pg_port: str, pg_db_name: str, pg_user_name: str, pg_password,
                 execution_time: str,
                 sftp_host: int, sftp_port: str, sftp_user_name: str, sftp_password: str, sftp_folder_path: str
                 ):

        self.sftp_folder_path = sftp_folder_path
        # remove last char if this is a slash
        if self.sftp_folder_path[-1:] == '/':
            self.sftp_folder_path = self.sftp_folder_path[:-1]
        self.sftp_password = sftp_password
        self.sftp_user_name = sftp_user_name
        self.sftp_port = sftp_port
        self.sftp_host = sftp_host
        self.execution_time = execution_time
        self.pg_password = pg_password
        self.pg_user_name = pg_user_name
        self.pg_db_name = pg_db_name
        self.pg_host = pg_host
        self.pg_port = pg_port

    def stream_pg_dump_to_sftp(self):
        try:
            # Establish SFTP connection
            transport = paramiko.Transport((self.sftp_host, self.sftp_port))
            transport.connect(username=self.sftp_user_name, password=self.sftp_password)
            sftp = paramiko.SFTPClient.from_transport(transport)

            # Build full file path: we use the name fo the day, so that only 7 files are kept and there is no need
            # to clean up ourselves (files are overwritten the next week)
            sftp_file_path = f"{self.sftp_folder_path}/{datetime.date.today().strftime('%A')}.gz"

            # Open a remote file for writing
            # TODO: given that the 'sftp.open()' works the same way as a regular file, we could make this script working for both cases
            with sftp.open(sftp_file_path, "wb") as remote_file:
                # Run pg_dump and capture output
                # Let's be honest, ChatGPT helped a lot on this ;-)
                process = subprocess.Popen(
                    ["pg_dump", "-U", self.pg_user_name, "-h", self.pg_host, "-p", self.pg_port, "-Fc", self.pg_db_name],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    env={"PGPASSWORD": self.pg_password}
                )

                # Compress the dump (the 'Fc' parameter of the pg_dump command is not very efficient)
                gzip_process = subprocess.Popen(["gzip"], stdin=process.stdout, stdout=subprocess.PIPE)

                # Stream gzip output directly to SFTP
                for chunk in iter(lambda: gzip_process.stdout.read(4096), b""):
                    remote_file.write(chunk)

                # Ensure the process completes
                process.stdout.close()
                process.wait()

                # Check for errors
                if process.returncode != 0:
                    error_message = process.stderr.read().decode()
                    logger.error(f"pg_dump failed: {error_message}")

            logger.info(f"Backup successfully uploaded to {self.sftp_folder_path}")

        except Exception as e:
            logger.error(f"Error: {e}")
            sys.exit(-1)

        finally:
            sftp.close()
            transport.close()


    def execute(self):
        logger.info("----- Initializing Postgres Dumper...")

        # Check if postgresql-client is installed
        try:
            result = subprocess.run(["pg_dump", "--version"], capture_output=True, text=True, check=True)
            logger.info(f"pg_dump version: {result.stdout.strip()}")
        except FileNotFoundError:
            logger.error("it seems that pg_dump is NOT installed, plase install it before running this script (apt install postgresql-client)")
            sys.exit(-1)

        if self.execution_time is None:
            # unit test case
            self.stream_pg_dump_to_sftp()
        else:
            # regular (prod) case
            schedule.every().day.at(self.execution_time).do(self.stream_pg_dump_to_sftp)
            while True:
                schedule.run_pending()
                time.sleep(1)


if __name__ == '__main__':
    level = logging.INFO
    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG
    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Periodically dumps Postgres DB to an SFTP server.')
    parser.add_argument('--pg_host', type=str, default='http://orthanc-db', help='Postgres hostname')
    parser.add_argument('--pg_port', type=str, default='5432', help='Postgres port number')
    parser.add_argument('--pg_db_name', type=str, default='postgres', help='Postgres database name')
    parser.add_argument('--pg_user_name', type=str, default='postgres', help='Postgres username')
    parser.add_argument('--pg_password', type=str, default='', help='Postgres password')
    parser.add_argument('--execution_time', type=str, default='01:30', help='Time for script execution (format: 23:30 or 23:30:14)')
    parser.add_argument('--sftp_host', type=str, default=None, help='sFTP server hostname')
    parser.add_argument('--sftp_port', type=str, default='22', help='sFTP server port number')
    parser.add_argument('--sftp_user_name', type=str, default=None, help='sFTP server user name')
    parser.add_argument('--sftp_password', type=str, default=None, help='sFTP server password')
    parser.add_argument('--sftp_folder_path', type=str, default=None, help='sFTP server folder path')
    args = parser.parse_args()

    pg_host = os.environ.get("PG_HOST", args.pg_host)
    pg_port = os.environ.get("PG_PORT", args.pg_port)
    pg_db_name = os.environ.get("PG_DB_NAME", args.pg_db_name)
    pg_user_name = os.environ.get("PG_USER_NAME", args.pg_user_name)
    pg_password = os.environ.get("PG_PASSWORD", args.pg_password)
    execution_time = os.environ.get("EXECUTION_TIME", args.execution_time)
    if execution_time is '':
        execution_time = None
    sftp_host = os.environ.get("SFTP_HOST", args.sftp_host)
    sftp_port = int(os.environ.get("SFTP_PORT", args.sftp_port))
    sftp_user_name = os.environ.get("SFTP_USER_NAME", args.sftp_user_name)
    sftp_password = os.environ.get("SFTP_PASSWORD", args.sftp_password)
    sftp_folder_path = os.environ.get("SFTP_FOLDER_PATH", args.sftp_folder_path)

    dumper = PostgresDumper(pg_host, pg_port, pg_db_name, pg_user_name, pg_password, execution_time,
                            sftp_host, sftp_port, sftp_user_name, sftp_password, sftp_folder_path)

    dumper.execute()



import argparse
import logging
from orthanc_api_client import OrthancApiClient
from typing import List
import zipfile
import tempfile
import os, time, sys
import multiprocessing
import queue
import threading

# examples:
# python orthanc_tools/orthanc_folder_importer.py --folder=./tests/stimuli --url=http://192.168.0.10:8042 --user=user --password=pwd --skip=.txt,.ini

# on a Windows system:
# python -m orthanc_tools.orthanc_folder_importer --url=https://pacs.orthanc.team/orthanc/ --api_key=**************** --folder_path=C:\\Orthanc --state_path=C:\\orthanc-migration\\status.txt --errors_path=C:\\orthanc-migration\\errors.txt --max_retries=2

logger = logging.getLogger(__name__)

class OrthancFolderImporter:
    '''
    Upload all the DICOM files contained in a folder (and its sub folders).
    It is a little bit smart:
    - There is a retry for every file and when it fails anyway, the file path is logged, but the script keeps working
    - Every sub folder uploaded (even with errors for some files) is logged, so that if the script is interrupted and
    restarted, it will restart from the last succeeded folder.
    - Zip files are unziped before upload
    '''
    def __init__(self,
                 api_client: OrthancApiClient,
                 folder_path: str,
                 errors_path: str,
                 state_path: str,
                 labels_list: List[str] = None,
                 max_retries: int = 8,
                 worker_threads_count: int = multiprocessing.cpu_count() - 1  # by default, use all CPUs but one for compression
                 ):
        self._api_client = api_client
        self._folder_path = folder_path
        self._labels_list = labels_list
        self._errors_path = errors_path # will contain the list of all the files path not correctly uploaded
        self._state_path = state_path # will contain the list of all the folders correctly uploaded

        self._worker_threads_count = worker_threads_count
        self._worker_threads = []
        self._messages = queue.Queue(maxsize=2*worker_threads_count)  # this is thread safe https://docs.python.org/3.5/library/queue.html#module-queue

        self._folders_uploaded = []

        if max_retries > 8:
            self._max_retries = 8
        else:
            self._max_retries = max_retries

        self._lock = threading.Lock()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def add_file_name_in_errors_log(self, file_path):
        with open(self._errors_path, "at") as f:
            f.write(file_path + "\n")

    def add_folder_path_in_state_file(self, folder_path):
        if self._state_path:
            with self._lock:
                with open(self._state_path, "at") as f:
                    f.write(folder_path + "\n")

    def upload_and_label(self, path_to_upload):
        """
        Upload the file if path_to_upload is a file path
        Recursively upload the content of the folder is path_to_upload is a folder path
        Then apply the labels on the study
        """

        # file path case
        if os.path.isfile(path_to_upload):

            # zip file case
            if "zip" in path_to_upload and zipfile.is_zipfile(path_to_upload):
                with tempfile.TemporaryDirectory() as tempDir:
                    with zipfile.ZipFile(path_to_upload, 'r') as z:
                        z.extractall(tempDir)
                    for path in os.listdir(tempDir):
                        full_path = os.path.join(tempDir, path)
                        self.upload_and_label(path_to_upload=full_path)
            else:
                retry_count = 0
                retry_delays = [5, 20, 60, 300, 900, 1800, 3600, 7200]

                while retry_count <= self._max_retries:
                    if retry_count >= 1:
                        delay = retry_delays[retry_count - 1]
                        logger.info(f"waiting {delay} seconds before retrying the upload of {path_to_upload}")
                        time.sleep(delay)
                    try:
                        # here, we should have only files (and no zip file)

                        # let's modify/filter the file if needed
                        with open(path_to_upload, 'rb') as f:
                            buffer = f.read()
                            buffer = self.process_dicom_file(buffer)

                        # filtering out case
                        if buffer is None:
                            logger.debug(f"File {path_to_upload} has been filtered out.")
                            return

                        # modification case: let's upload the file
                        logger.info(f"uploading {path_to_upload}")
                        instance_orthanc_ids = self._api_client.upload(buffer, ignore_errors=True)

                        if len(instance_orthanc_ids) == 0:
                            logger.error(f"File not uploaded: {path_to_upload}.")
                            self.add_file_name_in_errors_log(file_path=path_to_upload)
                            break
                        # we label for each instance, not at the end of the study, so that there is never an unlabeled image in Orthanc
                        if self._labels_list is not None:
                            study_orthanc_id = self._api_client.instances.get_parent_study_id(instance_orthanc_ids[0])
                            self._api_client.studies.add_labels(orthanc_id=study_orthanc_id, labels=self._labels_list)
                        break
                    except Exception as e:
                        if retry_count == self._max_retries:
                            logger.error(f"Error while uploading this file: {path_to_upload}. Exception: {str(e)}")
                            logger.error(f"too many attempts, logging the file name...")
                            self.add_file_name_in_errors_log(file_path=path_to_upload)
                            break
                        else:
                            retry_count += 1
                            logger.warning(f"Error while uploading this file, retrying...: {path_to_upload}. Exception: {str(e)}")
        # folder case
        elif os.path.isdir(path_to_upload):
            # this folder could have been processed in a previous run of the script
            if path_to_upload in self._folders_uploaded:
                logger.info(f"Folder {path_to_upload} already processed, skipping...")
                return

            # let's process this folder
            for path in os.listdir(path_to_upload):
                full_path = os.path.join(path_to_upload, path)
                self.upload_and_label(path_to_upload=full_path)

            # let's add this folder path in the processed ones:
            self.add_folder_path_in_state_file(path_to_upload)

    def process_dicom_file(self, file_content: bytes) -> bytes:
        '''
        This method is called just before the upload of the file to Orthanc
        By default, nothing is done, but one could want to apply some modifications on the data before upload
        or to filter out some files.
        To do so, this method should be overridden in a derived class.
        If the goal is to filter out the file, 'None' should be returned.

        file_content: content of the DICOM file, as a buffer of bytes
        output: a buffer of bytes (None to filter out the file)
        '''
        return file_content

    def _process_path(self, worker_id):
        logger.debug(f"Starting Processing thread {worker_id}")

        while True:
            path = self._messages.get()  # block until a message is available

            if path is None:  # sent by stop() to stop all worker threads
                self._messages.task_done()
                break

            # path is the full path of a file or a folder
            self.upload_and_label(path_to_upload=path)

            self._messages.task_done()  # tell the queue the item has been processed

        logger.debug("Processing thread stopped")

    def execute(self):
        # read state
        if self._state_path and os.path.isfile(self._state_path):
            with open(self._state_path, 'r') as file:
                lines = file.readlines()
                self._folders_uploaded = [line.strip() for line in lines]

        # create worker threads
        for thread_id in range(0, self._worker_threads_count):
            self._worker_threads.append(threading.Thread(
                target=self._process_path,
                name=f"Worker Thread {thread_id}",
                args=(thread_id, )
            ))

        # start threads
        for wt in self._worker_threads:
            wt.start()

        # let's browse the main folder to feed the message queue
        for path in os.listdir(path=self._folder_path):
            full_path = os.path.join(self._folder_path, path)
            self._messages.put(full_path) # if the queue is full, this will block until there's a free slot

        # let's wait for the completion of all threads
        self.stop()

        logger.info("End of upload!")

    def stop(self):
        logger.info("Waiting for Orthanc Folder Importer to complete each upload...")

        # post one 'empty' exit message per thread to unlock the threads from waiting on the process queue
        for i in range(0, self._worker_threads_count):
            self._messages.put(None)

        for t in self._worker_threads:
            t.join()

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Import the content of a folder in Orthanc')
    parser.add_argument('--url', type=str, default='http://localhost:8042', help='Orthanc url')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--api_key', type=str, default=None, help='Orthanc api-key')
    parser.add_argument('--folder_path', type=str, help='Folder to import, the one containing the DICOM files.')
    parser.add_argument('--labels_list', type=str, default=None, help='List of labels to apply to the uploaded studies, separated by a comma.')
    parser.add_argument('--errors_path', type=str, help='Path of the file which will contain the list of problematic files (not uploaded).')
    parser.add_argument('--state_path', type=str, help='Path of the file which will contain the list of all the folder correctly uploaded.')
    parser.add_argument('--max_retries', type=int, default=8, help='Maximum number of attempts for a file upload.')
    parser.add_argument('--worker_threads_count', type=int, default=1, help='Worker threads count')

    args = parser.parse_args()

    url = os.environ.get("ORTHANC_URL", args.url)
    user = os.environ.get("ORTHANC_USER", args.user)
    password = os.environ.get("ORTHANC_PWD", args.password)
    api_key = os.environ.get("ORTHANC_API_KEY", args.api_key)
    folder_path = os.environ.get("FOLDER_PATH", args.folder_path)
    labels_list = os.environ.get("LABELS_LIST", args.labels_list)
    errors_path = os.environ.get("ERRORS_PATH", args.errors_path)
    state_path = os.environ.get("STATE_PATH", args.state_path)
    max_retries = int(os.environ.get("MAX_RETRIES", str(args.max_retries)))
    worker_threads_count = int(os.environ.get("WORKER_THREADS_COUNT", str(args.worker_threads_count)))

    o = None
    if api_key is not None:
        o=OrthancApiClient(url, headers={"api-key": api_key})
    else:
        o=OrthancApiClient(url, user=user, pwd=password)

    importer = OrthancFolderImporter(
        api_client=o,
        folder_path=folder_path,
        labels_list=labels_list,
        errors_path=errors_path,
        state_path=state_path,
        max_retries=max_retries,
        worker_threads_count=worker_threads_count
    )

    importer.execute()



import os, time
import logging
import argparse
import tempfile
import datetime
from typing import List

from .helpers.scheduler import Scheduler
from orthanc_api_client import helpers
import logging
from orthanc_api_client import OrthancApiClient
import schedule

logger = logging.getLogger(__name__)

class OrthancSyncher:
    '''
    ## Goal
    The Syncher will ensure that all the studies/series/instances stored in Orthanc-1 are also
    stored in Orthanc-2 (and vice-versa).

    ## Context
    The modalities are configured to send the images to both Orthanc-1 and Orthanc-2, so that,
    if one of the 2 Orthanc is down/unreachable, the service is not interrupted.
    Then, there is a need for a nightly check: if a resource is missing from Orthanc-2 it has
    to be copied from Orthanc-1 to Orthanc-2 (and vice-versa).

    ## Challenge
    The `Comparator` tool does the exact same job... but it relies on the StudyDate for the queries.
    While this is great for a medical PACS, this won't work in research context, indeed, the
    stored resources could have a StudyDate in the far past...

    ## Trick
    The Syncher will query to `*` and filter on `LastUpdate` metadata. Then for each result, it
    will query the other Orthanc to check if the resource(s) is(are) also there (and act
    if needed).
    The latest value of the LastUpdate metadata is stored, so that the next check (next night),
    it will process only the recently modified resources.

    ## How it works
    There are 2 runs:
    First run to process the studies from Orthanc-1;
    Second run to process the studies from Orthanc-2.

    Each run will:
    - Get the list of all studies (by batches of 100)
    - For each study:
        - read the `LastUpdate` value
        - if `LastUpdate` < `LastProcessedLastUpdate`:
            - end of run.
        - if the study is not present in the other Orthanc:
            - send the study to the other Orthanc
        - else for each series of the study:
            - if the series is not present in the other Orthanc:
                - send the series to the other Orthanc
            - else for each instance of the series:
                - if the instance is not present in the other Orthanc:
                    - send the instance to the other Orthanc
    '''

    def __init__(self,
                 api_client_1: OrthancApiClient,
                 api_client_2: OrthancApiClient,
                 level: str = 'Series',
                 scheduler: Scheduler = None,
                 error_log_file_path: str = None,
                 persist_status_path: str = None,
                 run_till_last_update_1: datetime.datetime = None,
                 run_till_last_update_2: datetime.datetime = None,
                 execution_time: str = None,
                 execution_day: str = None,
                 orthanc_queries_batch_size: int = 100
                 ):

        if level not in ["Study", "Series", "Instance"]:
            raise RuntimeError("Invalid value for argument 'level'")

        self._api_client_1 = api_client_1
        self._api_client_2 = api_client_2
        self._scheduler = scheduler
        self._level = level
        self._error_log_file_path = error_log_file_path
        self._persist_status_path = persist_status_path

        self._current_run = "unknown"

        # first, we get the `run till` value from the file...
        if self._persist_status_path is not None:
            self._run_till_last_update_1, self._run_till_last_update_2 = self._read_status_from_file()

        # ...then, if there is nothing from the file, we try the arg of the constructor...
        else:
            if run_till_last_update_1 is not None:
                self._run_till_last_update_1 = run_till_last_update_1

            # ...finally, if nothing is in the args, we will process the entire content of Orthanc...
            else:
                self._run_till_last_update_1 = datetime.datetime(year=2025, month=1, day=1, hour=1, minute=1, second=1)

            if run_till_last_update_2 is not None:
                self._run_till_last_update_2 = run_till_last_update_2
            else:
                self._run_till_last_update_2 = datetime.datetime(year=2025, month=1, day=1, hour=1, minute=1, second=1)

        self._execution_time = execution_time
        self._execution_day = execution_day

        self._periodic_mode_enabled = False
        if self._execution_day is not None and self._execution_time is not None:
            self._periodic_mode_enabled = True

        self._batch_size = orthanc_queries_batch_size

    def _read_status_from_file(self):
        """
        The file should contain something like that:

        2025-05-20 14:01:58
        2025-05-19 15:44:36

        """
        try:
            with open(self._persist_status_path) as f:
                last_update_strings = f.read().splitlines()
        except (ValueError, FileNotFoundError):  # if can not read, use 1-1-1950
            logger.warning("Could not read LastUpdate values from file, running till at 01-01-1950")
            first_january_1950 = datetime.datetime(year=1950, month=1, day=1, hour=1, minute=1, second=1)

            with open(self._persist_status_path, 'w') as f:
                full_line = datetime.datetime.strftime(first_january_1950, "%Y-%m-%d %H:%M:%S") + '\n'
                f.writelines([full_line, full_line])
            return first_january_1950, first_january_1950

        last_update_1 = datetime.datetime.strptime(last_update_strings[0], "%Y-%m-%d %H:%M:%S")
        last_update_2 = datetime.datetime.strptime(last_update_strings[1], "%Y-%m-%d %H:%M:%S")

        logger.info(f"Orthanc-1 till value from file = {last_update_1}")
        logger.info(f"Orthanc-2 till value from file = {last_update_2}")
        return last_update_1, last_update_2

    def execute(self):

        # if one of the periodic mode parameters is missing, let's go for the regular mode
        if not self._periodic_mode_enabled:
            logger.info("----- Initializing Orthanc syncher (REGULAR mode, will run once)...")
            self._execute()
        else:
            logger.info("----- Initializing Orthanc syncher (PERIODIC mode, will run on a periodic basis)...")
            # 2 following lines allow to use a string as a method
            schedule_method = getattr(schedule.every(), self._execution_day)
            schedule_method.at(self._execution_time).do(self._execute)
            while True:
                schedule.run_pending()
                time.sleep(1)

    def _execute(self):
        # First run (Orthanc-1 studies are processed and pushed to Orthanc-2 if needed)

        self._current_run = "1 -> 2"
        logger.info(f"Starting run 1 ({self._current_run})...")

        self._run_till_last_update_1 = self.synch(
            orthanc_source=self._api_client_1,
            orthanc_destination=self._api_client_2,
            last_update_limit=self._run_till_last_update_1
        )

        if self._persist_status_path is not None:
            self.save_last_update_limit(self._run_till_last_update_1, 0)

        # Second run (Orthanc-2 studies are processed and pushed to Orthanc-1 if needed)

        self._current_run = "2 -> 1"
        logger.info(f"Starting run 2 ({self._current_run})...")

        self._run_till_last_update_2 = self.synch(
            orthanc_source=self._api_client_2,
            orthanc_destination=self._api_client_1,
            last_update_limit=self._run_till_last_update_2
        )
        if self._persist_status_path is not None:
            self.save_last_update_limit(self._run_till_last_update_2, 1)

    def synch(self, orthanc_source: OrthancApiClient, orthanc_destination: OrthancApiClient, last_update_limit: datetime.datetime) -> datetime.datetime:

        index=0
        while True:
            # Get a batch of studies
            studies = self.get_studies(orthanc_client=orthanc_source, batch_size=self._batch_size, index=index)

            logger.info(f"[{self._current_run}] Processing batch index {index}...")

            if index == 0:
                new_last_update_limit = studies[0].last_update

            for study in studies:
                if study.last_update >= last_update_limit:
                    self.compare_studies(orthanc_source, orthanc_destination, study.orthanc_id)
                else:
                    # the current study LastUpdate value is older than the most recent value checked during the last run, so it's over
                    # let's return the new_last_update_limit

                    logger.info(f"[{self._current_run}] Reached the 'LastUpdate' value already processed during last run, end of this run!")
                    return new_last_update_limit

            if len(studies) < self._batch_size:
                # if the actual batch size is less than the expected, all studies have been processed --> stop
                logger.info(f"[{self._current_run}] Processed all studies, end of this run!")
                return new_last_update_limit

            else:
                # if not, let's fetch the next batch
                index = index + self._batch_size

    def save_last_update_limit(self, new_last_update_value: datetime.datetime, index: int):
        try:
            # read file
            with open(self._persist_status_path, 'r') as f:
                last_update_strings = f.read().splitlines()

            # replace values
            last_update_strings[index] = datetime.datetime.strftime(new_last_update_value, "%Y-%m-%d %H:%M:%S")

            # write file
            with open(self._persist_status_path, 'w') as f:
                f.writelines(line + '\n' for line in last_update_strings)

        except (ValueError, FileNotFoundError):
            logger.error("Could not write LastUpdate values to file!")
            exit(0)

    def compare_studies(self, orthanc_source: OrthancApiClient, orthanc_destination: OrthancApiClient, orthanc_study_id):
        # check if study is in distant Orthanc
        if not orthanc_destination.studies.exists(orthanc_study_id):
            # if not there, transfer it
            instances_ids = orthanc_source.studies.get_instances_ids(orthanc_study_id)
            self.transfer_instances(orthanc_source, orthanc_destination, instances_ids)

        # study is there, let's check series/instances according to level
        elif self._level != "Study":
            source_series = orthanc_source.studies.get_series_ids(orthanc_study_id)
            for series_id in source_series:
                if not orthanc_destination.series.exists(series_id):
                    # this series is not in destination, transfer it
                    instances_ids = orthanc_source.series.get_instances_ids(series_id)
                    self.transfer_instances(orthanc_source, orthanc_destination, instances_ids)
                # series is there, let's check instances according to level
                elif self._level == "Instance":
                    source_instances = orthanc_source.series.get_instances_ids(series_id)
                    for instance_id in source_instances:
                        if not orthanc_destination.instances.exists(instance_id):
                            self.transfer_instances(orthanc_source, orthanc_destination, [instance_id])


    def transfer_instances(self, orthanc_source: OrthancApiClient, orthanc_destination: OrthancApiClient, instances_ids: List[str]):

        retry_count = 0
        retry_delays = [5, 20, 60, 300, 900, 1800, 3600, 7200]

        while retry_count <= 5:
            if retry_count >= 1:
                delay = retry_delays[retry_count - 1]
                logger.info(f"waiting {delay} seconds before retrying transfer for resource from study {orthanc_source.instances.get_parent_study_id(instances_ids[0])}")
                time.sleep(delay)

            try:
                for instance_id in instances_ids:
                    data = orthanc_source.instances.get_file(instance_id)
                    orthanc_destination.upload(data)

                break

            except Exception as e:
                if retry_count == 5:
                    logger.error(f"Error while transfering a resource from this study: {orthanc_source.instances.get_parent_study_id(instances_ids[0])}. Exception: {str(e)}")
                    exit(0)
                else:
                    retry_count += 1
                    logger.warning(f"Error while transfering a resource from this study: {orthanc_source.instances.get_parent_study_id(instances_ids[0])}. Exception: {str(e)}")

    def get_studies(self, orthanc_client: OrthancApiClient, batch_size: int, index: int):
        return orthanc_client.studies.find(
            query={},
            limit = batch_size,
            since = index,
            order_by = [
                {
                    "Type": "Metadata",
                    "Key": "LastUpdate",
                    "Direction": "DESC"
                }
            ]
        )


if __name__ == '__main__':
    level = logging.INFO

    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    parser = argparse.ArgumentParser(description='Ensure that all the studies/series/instances stored in Orthanc-1 are also stored in Orthanc-2.')
    parser.add_argument('--url_1', type=str, default='http://localhost:8042', help='Orthanc-1 url')
    parser.add_argument('--user_1', type=str, default=None, help='Orthanc-1 user name')
    parser.add_argument('--password_1', type=str, default=None, help='Orthanc-1 password')
    parser.add_argument('--api_key_1', type=str, default=None, help='Orthanc-1 api-key')
    parser.add_argument('--url_2', type=str, default='http://localhost:8042', help='Orthanc-2 url')
    parser.add_argument('--user_2', type=str, default=None, help='Orthanc-2 user name')
    parser.add_argument('--password_2', type=str, default=None, help='Orthanc-2 password')
    parser.add_argument('--api_key_2', type=str, default=None, help='Orthanc-2 api-key')
    parser.add_argument('--level', type=str, default='Series', help='Compare resources up to Study/Series/Instance level')
    parser.add_argument('--error_log_file_path', type=str, default='/errors.log', help='Path to the file to write errors log')
    parser.add_argument('--persist_status_path', type=str, default='/status.txt', help='Path to the file to write the status')

    parser.add_argument('--execution_time', type=str, default=None,
                        help='Enables periodic mode. The time when the periodic run will start (format: 23:30 or 23:30:14).')
    parser.add_argument('--execution_day', type=str, default=None,
                        help='Enables periodic mode. The day when the periodic run will start. All days of the week are valid values (lowercase), \'day\' is also accepted for everyday runs.')

    Scheduler.add_parser_arguments(parser)

    args = parser.parse_args()

    url_1 = os.environ.get("ORTHANC_1_URL", args.url_1)
    user_1 = os.environ.get("ORTHANC_1_USER", args.user_1)
    password_1 = os.environ.get("ORTHANC_1_PWD", args.password_1)
    api_key_1 = os.environ.get("ORTHANC_1_API_KEY", args.api_key_1)
    
    url_2 = os.environ.get("ORTHANC_2_URL", args.url_2)
    user_2 = os.environ.get("ORTHANC_2_USER", args.user_2)
    password_2 = os.environ.get("ORTHANC_2_PWD", args.password_2)
    api_key_2 = os.environ.get("ORTHANC_2_API_KEY", args.api_key_2)
    
    level = os.environ.get("LEVEL", args.level)

    error_log_file_path = os.environ.get("ERROR_LOG_FILE_PATH", args.error_log_file_path)
    persist_status_path = os.environ.get("PERSIST_STATUS_PATH", args.persist_status_path)
    
    execution_time = os.environ.get("EXECUTION_TIME", args.execution_time)
    execution_day = os.environ.get("EXECUTION_DAY", args.execution_day)

    scheduler = Scheduler.create_from_args_and_env_var(args)

    api_client_1 = None
    if api_key_1 is not None:
        api_client_1 = OrthancApiClient(url_1, headers={"api-key" : api_key_1})
    else:
        api_client_1 = OrthancApiClient(url_1, user=user_1, pwd=password_1)
        
    api_client_2 = None
    if api_key_2 is not None:
        api_client_2 = OrthancApiClient(url_2, headers={"api-key" : api_key_2})
    else:
        api_client_2 = OrthancApiClient(url_2, user=user_2, pwd=password_2)



    syncher = OrthancSyncher(
        api_client_1=api_client_1,
        api_client_2=api_client_2,
        level=level,
        scheduler=scheduler,
        error_log_file_path=error_log_file_path,
        persist_status_path=persist_status_path,
        execution_time=execution_time,
        execution_day=execution_day
    )

    syncher.execute()



